/*
    CFS0.11 filesystem implementation for FUSE
    Copyright (C) 2005-2017 Kajtar Zsolt (Soci/Singular) <soci@c64.rulez.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#define _LARGEFILE64_SOURCE

#define FUSE_USE_VERSION 26

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <locale.h>
#include <wchar.h>

#define offset_t unsigned
#ifdef MB_CUR_MAX
#undef MB_CUR_MAX
#endif
#define MB_CUR_MAX 6

#define FTYPE_DEL 0
#define FTYPE_PRG 1
#define FTYPE_REL 2
#define FTYPE_DIR 3
#define FTYPE_LNK 4
#define FTYPE_MASK 7

static pthread_mutex_t big_lock = PTHREAD_MUTEX_INITIALIZER;
static offset_t partition_table; //constant
static unsigned int geometry_cylinder; //constant
static unsigned int geometry_head; //constant
static unsigned int geometry_sector; //constant
static time_t epoch_time = 0; //constant
static struct stat stat_default; //constant

static int f1 = -1; //constant

typedef unsigned char c64_t;

typedef struct {
    int valid;
    int bam;
    c64_t name[17];
    offset_t rootdir;
    offset_t deldir;
    offset_t int start;
    offset_t int end;
    off_t size;
    off_t free;
} part_t;

typedef struct file_s {
    size_t refcount;
    offset_t treeroot;
    offset_t lastblock, lasttree;
    part_t *partition;
    offset_t dir_sector;
    int dir_offset;
    struct file_s *next;
} file_t;

static file_t *files_head = NULL;

typedef struct {
    mode_t attr;
    off_t size;
    part_t *partition;
    int file_offset;
    offset_t dir_sector;
    int empty_offset;
    offset_t empty_sector;
    offset_t this_dir;
    offset_t end_of_this_dir;
    offset_t root;
    size_t sectors;
    c64_t name[16 + 1];
    c64_t type[3 + 1];
    unsigned char date[4];
} resolv_t;

typedef struct {
    offset_t offset;
    int dirty;
    int next;
    int fault;
    unsigned char buffer[512];
} buffer_t;

static part_t partitions[16]; //constant

static int to_petscii(c64_t *s2, const char *s, ssize_t len, size_t max) {
    mbstate_t ps;
    memset(&ps, 0, sizeof(mbstate_t));
    wchar_t w;
    ssize_t l;
    while (*s && len > 0 && max) {
        if (*s == '\\' && s[1] == 'u') {
            int j;
            w = 0; l = 2;
            for (j = 12; j >= 0; j -= 4, l++) {
                if (s[l] >= '0' && s[l] <= '9') w |= (s[l] - '0') << j;
                else if (s[l] >= 'a' && s[l] <= 'f') w |= (s[l] - 'a' + 10) << j;
                else {
                    w = *s;
                    l = 1;
                    break;
                }
            }
        } else if (*s >= 0x20 && *s <= 0x7e) {
            w = *s;
            l = 1;
        } else {
            l = mbrtowc(&w, s, MB_CUR_MAX, &ps);
        }
        if (l < 0) return 1;
        s += l; len -= l;
        if (w >= 0x20 && w <= 0x40) *s2 = w;
        else if (w >= 0x41 && w <= 0x5a) *s2 = w + 0x80;
        else if (w >= 0x61 && w <= 0x7a) *s2 = w - 0x20;
        else if (w >= 0xe100 && w <= 0xe1ff) *s2 = w - 0xe100;
        else switch (w) {
        case 0x2215: *s2 = 0x2f; break;
        case 0x5b:   *s2 = 0x5b; break;
        case 0xa3:   *s2 = 0x5c; break;
        case 0x5d:   *s2 = 0x5d; break;
        case 0x2191: *s2 = 0x5e; break;
        case 0x2190: *s2 = 0x5f; break;
        case 0xa0:   *s2 = 0xa0; break;
        case 0x258c: *s2 = 0xa1; break;
        case 0x2584: *s2 = 0xa2; break;
        case 0x2594: *s2 = 0xa3; break;
        case 0x5f:   *s2 = 0xa4; break;
        case 0x258f: *s2 = 0xa5; break;
        case 0x2592: *s2 = 0xa6; break;
        case 0x251c: *s2 = 0xab; break;
        case 0x2597: *s2 = 0xac; break;
        case 0x2514: *s2 = 0xad; break;
        case 0x2510: *s2 = 0xae; break;
        case 0x2582: *s2 = 0xaf; break;
        case 0x250c: *s2 = 0xb0; break;
        case 0x2534: *s2 = 0xb1; break;
        case 0x252c: *s2 = 0xb2; break;
        case 0x2524: *s2 = 0xb3; break;
        case 0x258e: *s2 = 0xb4; break;
        case 0x258d: *s2 = 0xb5; break;
        case 0x2583: *s2 = 0xb9; break;
        case 0x2713: *s2 = 0xba; break;
        case 0x2596: *s2 = 0xbb; break;
        case 0x259d: *s2 = 0xbc; break;
        case 0x2518: *s2 = 0xbd; break;
        case 0x2598: *s2 = 0xbe; break;
        case 0x259a: *s2 = 0xbf; break;
        case 0x2500: *s2 = 0xc0; break;
        case 0x253c: *s2 = 0xdb; break;
        case 0x2502: *s2 = 0xdd; break;
        default: return 1;
        }
        s2++; max--;
    }
    *s2 = 0;
    return 0;
}

static int to_ascii(char *st2, const c64_t *s, int link) {
    wchar_t s2;
    mbstate_t ps;
    size_t i = 0, l;
    memset(&ps, 0, sizeof(mbstate_t));
    while (*s) {
        char mb[MB_CUR_MAX];
        if (*s >= 0x20 && *s <= 0x40 && (*s != 0x2f || link)) s2 = *s;
        else if (*s >= 0x41 && *s <= 0x5a) s2 = *s + 0x20;
        else if (*s >= 0xc1 && *s <= 0xda) s2 = *s - 0x80;
        else switch (*s) {
        case 0x2f: s2 = 0x2215; break;
        case 0x5b: s2 = 0x5b; break;
        case 0x5c: s2 = 0xa3; break;
        case 0x5d: s2 = 0x5d; break;
        case 0x5e: s2 = 0x2191; break;
        case 0x5f: s2 = 0x2190; break;
        case 0xa0: s2 = 0xa0; break;
        case 0xa1: s2 = 0x258c; break;
        case 0xa2: s2 = 0x2584; break;
        case 0xa3: s2 = 0x2594; break;
        case 0xa4: s2 = 0x5f; break;
        case 0xa5: s2 = 0x258f; break;
        case 0xa6: s2 = 0x2592; break;
        case 0xab: s2 = 0x251c; break;
        case 0xac: s2 = 0x2597; break;
        case 0xad: s2 = 0x2514; break;
        case 0xae: s2 = 0x2510; break;
        case 0xaf: s2 = 0x2582; break;
        case 0xb0: s2 = 0x250c; break;
        case 0xb1: s2 = 0x2534; break;
        case 0xb2: s2 = 0x252c; break;
        case 0xb3: s2 = 0x2524; break;
        case 0xb4: s2 = 0x258e; break;
        case 0xb5: s2 = 0x258d; break;
        case 0xb9: s2 = 0x2583; break;
        case 0xba: s2 = 0x2713; break;
        case 0xbb: s2 = 0x2596; break;
        case 0xbc: s2 = 0x259d; break;
        case 0xbd: s2 = 0x2518; break;
        case 0xbe: s2 = 0x2598; break;
        case 0xbf: s2 = 0x259a; break;
        case 0xc0: s2 = 0x2500; break;
        case 0xdb: s2 = 0x253c; break;
        case 0xdd: s2 = 0x2502; break;
        default: s2 = 0xe100 + *s;
        }
        s++;
        if (s2 >= 0x20 && s2 <= 0x7e) {
            st2[i++] = s2;
            continue;
        }
        l = wcrtomb(mb, s2, &ps);
        if (l == (size_t)-1) {
            l = sprintf(mb, "\\u%04x", (unsigned int)s2);
        }
        memcpy(st2 + i, mb, l);
        i += l;
    }
    st2[i] = 0;
    return i;
}

static const offset_t null_sector = (offset_t)-1;
static offset_t sector2offset(const unsigned char *ez) {
    offset_t cylinder, head, sector;
    head = ez[0] & 0x0f;
    cylinder = (ez[1] << 8) + ez[2];
    sector = ez[3];
    if (ez[0] & 0x40) {
        return sector + (cylinder << 8) + (head << 24);
    }
    if (sector == 0 || sector > geometry_sector || head >= geometry_head || cylinder > geometry_cylinder) return null_sector;
    return (cylinder * geometry_head + head) * geometry_sector + sector - 1;
}

static void offset2sector(unsigned char *ide, offset_t ezt) {
    if (geometry_head & 0x40) {
        ide[0] = ((ezt >> 24) & 0x0f) | 0x40;
        ide[1] = ezt >> 16;
        ide[2] = ezt >> 8;
        ide[3] = ezt;
        return;
    }
    ide[3] = (ezt % geometry_sector) + 1;
    ezt /= geometry_sector;
    ide[0] = ezt % geometry_head;
    ezt /= geometry_head;
    ide[1] = ezt >> 8;
    ide[2] = ezt;
}

static void offset2sector_insert(unsigned char *ide, offset_t ezt) {
    unsigned char old = ide[0] & 0xb0;
    offset2sector(ide, ezt);
    ide[0] = (ide[0] & 0x4f) | old;
}

static int write_sector(const unsigned char *ezt, offset_t ide) {
    if (ide == null_sector) return -EIO;
//    fprintf(stderr, "WRITE (%u)\n", ide);
#ifdef __CYGWIN__
    if (lseek(f1, (off_t)ide << 9, SEEK_SET) < 0 || write(f1, ezt, 512) < 0) return errno;
#else
    if (pwrite(f1, ezt, 512, (off_t)ide << 9) < 0) return errno;
#endif
    return 0;
}

static int read_sector(unsigned char *ezt, offset_t ide) {
    ssize_t ret;
    if (ide == null_sector) {
        memset(ezt, 0, 512);
        return -EIO;
    }
//    fprintf(stderr, "READ (%u)\n", ide);
#ifdef __CYGWIN__
    if (lseek(f1, (off_t)ide << 9, SEEK_SET) == (off_t)-1) {
        memset(ezt, 0, 512);
        ret = -1;
    } else ret = read(f1, ezt, 512);
#else
    ret = pread(f1, ezt, 512, (off_t)ide << 9);
#endif
    if (ret < 0) return errno;
    if (ret < 512) memset(ezt + ret, 0, 512 - ret);
    return 0;
}

#define CACHE_SIZE 32
static buffer_t buffers[CACHE_SIZE];
static int buffer_head, buffer_tail;

static buffer_t *buffer_find(offset_t ide) {
    int i, j = buffer_head;
    buffer_t *buffer;
    for (i = 0; i < CACHE_SIZE; i++) {
        buffer = &buffers[j];
        if (buffer->offset == ide) {
            if (j == buffer_head) break;
            buffers[buffer_tail].next = buffer->next;
            buffer->next = buffer_head;
            buffer_head = j;
            break;
        }
        buffer_tail = j;
        j = buffer->next;
    }
    return buffer;
}

static int buffer_flush(buffer_t *buffer) {
    int ret = 0;
    if (buffer->dirty) {
        ret = write_sector(buffer->buffer, buffer->offset);
        buffer->dirty = 0;
    }
    buffer->next = buffer_head;
    buffer_head = buffer_tail;
    buffer->fault = 0;
    buffer->offset = -1;
    return ret;
}

static buffer_t *buffer_new(offset_t ide) {
    buffer_t *buffer = buffer_find(ide);
    if (buffer->offset != ide) buffer_flush(buffer);
    memset(buffer->buffer, 0, 512);
    buffer->fault = 0;
    buffer->offset = ide;
    buffer->dirty = 1;
    return buffer;
}

static buffer_t *buffer_read(offset_t ide) {
    buffer_t *buffer = buffer_find(ide);
    if (buffer->offset != ide) {
        buffer_flush(buffer);
        buffer->fault = read_sector(buffer->buffer, ide);
        buffer->offset = ide;
    }
    return buffer;
}

static void buffer_done(buffer_t *buffer) {
    (void) buffer;
}

static int buffers_sync(int ret) {
    int i, j = 0, ret2;
    int resort[CACHE_SIZE];
    for (i = 0; i < CACHE_SIZE; i++) {
        if (buffers[i].dirty) {
            resort[j++] = i;
        }
    }
    while (j--) {
        buffer_t *buffer = &buffers[resort[j]];
        ret2 = write_sector(buffer->buffer, buffer->offset);
        if (ret2) ret = ret2;
        buffer->dirty = 0;
    }
    return ret;
}

static void fill_in_time(unsigned char *sbuff, time_t ee) {
    struct tm ido2;

    localtime_r(&ee, &ido2);

    ido2.tm_mon++;
    sbuff[0] = (ido2.tm_sec % 60) | ((ido2.tm_mon << 4) & 0xc0);
    sbuff[1] = ido2.tm_min | (ido2.tm_mon << 6);
    sbuff[2] = ((ido2.tm_year - 80) % 100) | ((ido2.tm_hour << 3) & 0xc0);
    sbuff[3] = ido2.tm_mday | (ido2.tm_hour << 5);
}

static const char corr[16] = {
    0, 0, 1, 5, 6, 8, 9, 11, 12, 13, 15, 16, 18, 0, 0, 0
};

static void read_in_time(struct stat *at, const unsigned char *t) {
    static unsigned char last[4];
    static time_t last_time;

    if (memcmp(last, t, 4) != 0) {
        memcpy(last, t, 4);

        if (t[3] & 0x1f) {
            int i = (((t[0] >> 4) & 0x0c) | (t[1] >> 6));
            time_t new_time = t[2] & 0x3f;
            new_time = new_time * 1461 / 4;
            new_time -= (t[2] & 3) == 0 && i < 3;
            new_time += i * 32 - corr[i];
            new_time += t[3] & 0x1f;
            new_time *= 24;
            new_time += ((t[2] >> 3) & 0x18) | (t[3] >> 5);
            new_time *= 60;
            new_time += t[1] & 0x3f;
            new_time *= 60;
            new_time += t[0] & 0x3f;
            last_time = new_time + epoch_time;
        } else {
            last_time = stat_default.st_mtime;
        }
    }
    at->st_mtime = last_time;
    at->st_atime = last_time;
    at->st_ctime = last_time;
}

static int find_free_sector(part_t *partition, offset_t *st, int data) {
    offset_t itt, fraction, bamm, start = *st;
    unsigned int b;
    buffer_t *bam;

    if (start < partition->start || start >= partition->end) start = partition->start;

    bamm = partition->bam - partition->start;
    fraction = (start - partition->start) & 0xfff;

back:
    itt = start - fraction;
    b = data ? 128 : 2;
    if (data && fraction > b) b = fraction;
    do {
        bam = buffer_read(itt + bamm);
        if (!bam->fault) {
            for (; b < 4096; b++) {
                unsigned char f = bam->buffer[b >> 3];
                if (f == 0) {
                    b |= 7;
                    continue;
                }
                if ((f << (b & 7)) & 0x80) goto ok;
            }
        }
        b = data ? 128 : 2;
        buffer_done(bam);
        itt += 4096;
    } while (itt < partition->end);

    itt = start - fraction;
    do {
        bam = buffer_read(itt + bamm);
        if (!bam->fault) {
            for (b = data ? 128 : 2; b < 4096; b++) {
                unsigned char f = bam->buffer[b >> 3];
                if (f == 0) {
                    b |= 7;
                    continue;
                }
                if ((f << (b & 7)) & 0x80) goto ok;
            }
        }
        buffer_done(bam);
        if (itt < 4096) break;
        itt -= 4096;
    } while (itt > partition->start);
    if (data) {
        data = 0;
        goto back;
    }
    return -ENOSPC;
ok:
    bam->buffer[b >> 3] &= ~(128 >> (b & 7));
    bam->dirty = 1;
    buffer_done(bam);
    if (partition->free) partition->free--;

    *st = itt + b;
    return 0;
}

static void free_this_sector(part_t *partition, offset_t thiss) {
    offset_t itt;
    unsigned int b;
    buffer_t *bam;

    if (thiss < partition->start) return;
    if (thiss >= partition->end) return;

    b = (thiss - partition->start) & 0xfff;

    if (b < 2) return;

    itt = thiss - b;
    itt += partition->bam - partition->start;

    bam = buffer_read(itt);
    if (!bam->fault) {
        if (~bam->buffer[b >> 3] & (128 >> (b & 7))) {
            bam->buffer[b >> 3] |= 128 >> (b & 7);
            bam->dirty = 1;
            partition->free++;
        }
    }
    buffer_done(bam);
}

static unsigned int get_entry_size(const unsigned char *entry) {
    unsigned int size = entry[0x10] + (entry[0x11] << 8) + (entry[0x12] << 16);
    return ((entry[0x18] & FTYPE_MASK) == FTYPE_REL) ? size : (size + (entry[0x13] << 24));
}

static void set_entry_size(unsigned char *entry, unsigned int size) {
    entry[0x10] = size;
    entry[0x11] = size >> 8;
    entry[0x12] = size >> 16;
    if ((entry[0x18] & FTYPE_MASK) != FTYPE_REL) entry[0x13] = size >> 24;
}

static void fill_in_resolv(resolv_t *at, offset_t itt, int i, const unsigned char *entry) {
    unsigned char attr = entry[0x18];
    at->dir_sector = itt;
    at->file_offset = i;
    at->size = get_entry_size(entry);
    memcpy(at->date, entry + 0x1c, 4);
    at->attr = (((attr >> 3) & 7) * 0111) & 0755;
    switch (attr & FTYPE_MASK) {
    case FTYPE_DIR: at->attr |= S_IFDIR; at->size = 512; break;
    case FTYPE_LNK: at->attr |= S_IFLNK; break;
    default: at->attr |= S_IFREG;
    }
    at->root = sector2offset(entry + 0x14);
}

static offset_t get_dir_next(const unsigned char *buffer) {
    unsigned char next_sector[4], c;
    const unsigned char *entry = buffer + 0x200;
    int i;

    for (i = 0; i < 4; i++) {
        entry -= 0x80;
        c = (entry[0x14] << 2) & 0xc0;
        c |= entry[0x34] & 0x30;
        c |= (entry[0x54] >> 2) & 0x0c;
        next_sector[i] = c | ((entry[0x74] >> 4) & 0x03);
    }
    return sector2offset(next_sector);
}

static void set_dir_next(unsigned char *buffer, offset_t ezt) {
    unsigned char next_sector[4];
    unsigned char *entry = buffer + 0x200;
    int i, j;

    if (ezt == null_sector) {
        memset(next_sector, 0, 4);
    } else {
        offset2sector(next_sector, ezt);
    }
    for (j = 0; j < 4; j++) {
        unsigned int c = next_sector[j] << 4;
        for (i = 0; i < 4; i++) {
            entry -= 0x20;
            entry[0x14] = (entry[0x14] & ~0x30) | (c & 0x30);
            c >>= 2;
        }
    }
}

#define RESOLV_DIR 1
#define RESOLV_FILE 2
#define RESOLV_LINK 3
#define RESOLV_ALL 4
static int resolv_path(resolv_t *at, const char *path, int smode) {
    offset_t itt, last;
    const char *p, *p2;
    int i, j, k;
    int cmode;
    int vesszo;
    c64_t name[16 + 1];
    c64_t type[3 + 1];

    p = path;
    at->date[0] = 0;
    at->date[1] = 64;
    at->date[2] = 0;
    at->date[3] = 1;
    at->root = partition_table;
    at->sectors = 0;
    at->empty_sector = null_sector;
    at->partition = NULL;
    at->attr = S_IFDIR | 0555;
    at->size = 512;
    at->file_offset = -1;
    while (*p == '/') p++;
    if (!*p) return 0;

    if (p[0] < '0' || p[0] > '9') return 1;
    if (p[1] < '0' || p[1] > '9') return 1;
    if (p[2] != ' ') return 1;
    i = p[0] * 10 + p[1] - 529;
    if (i < 0 || i > 15) return 1;
    if (!partitions[i].valid) return 1;
    itt = partitions[i].rootdir;
    at->root = itt;
    at->attr = S_IFDIR | 0755;
    while (*p != '/' && *p) p++;
    if (!*p) return 0;
    at->partition = &partitions[i];
    at->type[0] = at->name[0] = 0;

nextdir:
    at->sectors = 0;
    last = null_sector;
    at->empty_sector = null_sector;
    at->this_dir = itt;
    while (*p == '/') p++;
    p2 = p;
    while (*p2 != '/' && *p2) p2++;
    cmode = *p2 ? RESOLV_DIR : smode;

    j = 0;
    while (p[j] && p[j] != '/' && p[j] != ',') j++;
    if (to_petscii(name, p, j, 16)) return 1;
    vesszo = (p[j] == ',');
    if (vesszo) {
        size_t j2 = ++j;
        while (p[j] && p[j] != '/') j++;
        j2 = j - j2;
        if (to_petscii(type, p + j - j2, j2, 3)) return 1;
        if (!type[1]) {
            const char *t;
            switch (type[0]) {
            case 'P': t = "PRG"; break;
            case 'S': t = "SEQ"; break;
            case 'U': t = "USR"; break;
            default: return 1;
            }
            memcpy(type, t, sizeof(type));
        }
    } else memcpy(type, cmode == RESOLV_DIR ? "DIR" : cmode == RESOLV_LINK ? "LNK" : "PRG", sizeof(type));

    while (itt != null_sector) {
        buffer_t *sbuff = buffer_read(itt);
        if (sbuff->fault) {
            int ret = sbuff->fault;
            buffer_done(sbuff);
            return ret;
        }
        at->sectors++;

        for (i = 0; i < 512; i += 32) {
            unsigned char *entry = sbuff->buffer + i;
            unsigned char attr = entry[0x18], ftype = (attr & FTYPE_MASK);
            if ((attr & 0x80) == 0 || ftype == FTYPE_DEL) {
                if ((attr & 0x87) == 0 && at->empty_sector == null_sector) {
                    at->empty_offset = i;
                    at->empty_sector = itt;
                }
                continue;
            }
            switch (cmode) {
            case RESOLV_LINK:
            case RESOLV_FILE: if (ftype != FTYPE_PRG && ftype != FTYPE_REL && ftype != FTYPE_LNK) continue; else break;
            case RESOLV_DIR: if (ftype != FTYPE_DIR) continue; else break;
            default: break;
            }
            for (k = 0; k < 16; k++) {
                if (!name[k] || name[k] != entry[k]) break;
            }
            if (k < 16 && name[k] != entry[k]) continue;

            if (vesszo || cmode != RESOLV_ALL || (ftype != FTYPE_DIR && ftype != FTYPE_LNK)) {
                unsigned char *etype = entry + 0x19;
                for (k = 0; k < 3; k++) {
                    if (!type[k] || type[k] != etype[k]) break;
                }
                if (k < 3 && type[k] != etype[k]) continue;
            }
            
            if (p[j] == '/') {
                p += j;
                itt = sector2offset(entry + 0x14);
                buffer_done(sbuff);
                goto nextdir;
            }
            if (!p[j]) {
                fill_in_resolv(at, itt, i, entry);
                memcpy(at->name, name, sizeof(at->name));
                memcpy(at->type, vesszo ? type : (entry + 0x19), sizeof(at->type));
                at->type[3] = 0;
                at->end_of_this_dir = last;
                if (cmode != RESOLV_ALL || ftype == FTYPE_DIR) {
                    buffer_done(sbuff);
                    return 0;
                }
                cmode = RESOLV_DIR;
            }
        }
        last = itt;
        itt = get_dir_next(sbuff->buffer);
        buffer_done(sbuff);
    }
    if (at->name[0]) return 0;
    memcpy(at->name, name, sizeof(name));
    memcpy(at->type, type, sizeof(type));
    at->end_of_this_dir = last;
    return 1;
}

static int cfs011_fgetattr(const char *path, struct stat *stbuf, struct fuse_file_info *fi) {
    resolv_t at;
    int ret = 0;

    *stbuf = stat_default;
    stbuf->st_uid = fuse_get_context()->uid;
    stbuf->st_gid = fuse_get_context()->gid;

    pthread_mutex_lock(&big_lock);
    if (fi == NULL) {
        ret = resolv_path(&at, path, RESOLV_ALL);
        if (ret > 0) ret = -ENOENT;
    } else {
        file_t *file = (file_t *) (uintptr_t) fi->fh;
        buffer_t *sbuff = buffer_read(file->dir_sector);
        ret = sbuff->fault;
        fill_in_resolv(&at, file->dir_sector, file->dir_offset, sbuff->buffer + file->dir_offset);
        buffer_done(sbuff);
    }
    pthread_mutex_unlock(&big_lock);

    stbuf->st_mode = at.attr;
    stbuf->st_size = at.size;
    stbuf->st_ino = at.root;
    stbuf->st_blocks = (stbuf->st_size / 512) + ((stbuf->st_size + 65535) / 512 / 128) + 1;

    if (at.partition == NULL) {
        stbuf->st_nlink = 3;
	return ret;
    }
    read_in_time(stbuf, at.date);

    stbuf->st_nlink = (stbuf->st_mode & S_IFDIR) ? 3 : 1;
    return ret;
}

static int cfs011_getattr(const char *path, struct stat *stbuf) {
    return cfs011_fgetattr(path, stbuf, NULL);
}

static int cfs011_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                       off_t offset, struct fuse_file_info *fi)
{
    char name[(16 + 4) * MB_CUR_MAX + 1];
    int i, j;
    offset_t itt;
    struct stat st;
    resolv_t at;
    int ret;

    (void) offset;
    (void) fi;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_DIR);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    st = stat_default;
    if (at.root == partition_table) {
        st.st_mode = S_IFDIR;
        st.st_ino = partition_table;
        filler(buf, ".", &st, 0);
        filler(buf, "..", &st, 0);
        for (i = 0; i < 16; i++) {
            part_t *partition = &partitions[i];
	    if (!partition->valid) continue; // valid?
            name[0] = (i >= 9) ? '1' : '0';
            name[1] = ((i >= 9) ? '\'' : '1') + i;
            name[2] = ' ';
            to_ascii(name + 3, partition->name, 0);
            st.st_ino = partition->rootdir;
	    filler(buf, name, &st, 0);
        }
        goto fault;
    }
    itt = at.root;
    while (itt != null_sector) {
        buffer_t *sbuff = buffer_read(itt);
        if (sbuff->fault) {
            ret = sbuff->fault;
            buffer_done(sbuff);
            break;
        }
        for (i = 0; i < 512; i += 32) {
            unsigned char *entry = sbuff->buffer + i, tmp;
            switch (entry[0x18] & (FTYPE_MASK | 0x80)) {
            case 0x80 | FTYPE_PRG:
            case 0x80 | FTYPE_REL: st.st_mode = S_IFREG; break;
            case 0x80 | FTYPE_DIR: st.st_mode = S_IFDIR; break;
            case 0x80 | FTYPE_LNK: st.st_mode = S_IFLNK; break;
            case FTYPE_DIR:
                st.st_mode = S_IFDIR;
                st.st_ino = sector2offset(entry + 0x10);
                filler(buf, ".", &st, 0);
                st.st_ino = (at.partition == NULL) ? partition_table : sector2offset(entry + 0x14);
                filler(buf, "..", &st, 0);
                continue;
            default: continue;
            }
            tmp = entry[16]; entry[16] = 0;
            j = to_ascii(name, entry, 0);
            entry[16] = tmp;
            if (st.st_mode != S_IFDIR) {
                name[j++] = ',';
                tmp = entry[0x1c]; entry[0x1c] = 0;
                to_ascii(name + j, entry + 0x19, 0);
                entry[0x1c] = tmp;
            }
            st.st_ino = sector2offset(entry + 0x14);
            filler(buf, name, &st, 0);
        }
        itt = get_dir_next(sbuff->buffer);
        buffer_done(sbuff);
    }
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static file_t *file_lookup(const resolv_t *at) {
    file_t *file;
    for (file = files_head; file != NULL; file = file->next) {
        if (file->partition == at->partition &&
                file->dir_sector == at->dir_sector &&
                file->dir_offset == at->file_offset) return file;
    }
    return NULL;
}

static file_t *file_open(const resolv_t *at) {
    file_t *file = file_lookup(at);
    
    if (file == NULL) {
        file = (file_t*)malloc(sizeof *file);
        if (file == NULL) return NULL;
        file->refcount = 0;
        file->treeroot = at->root;
        file->lastblock = at->root;
        file->lasttree = at->dir_sector;
        file->partition = at->partition;
        file->dir_sector = at->dir_sector;
        file->dir_offset = at->file_offset;
    }
    file->refcount++;
    return file;
}

static int cfs011_open(const char *path, struct fuse_file_info *fi) {
    resolv_t at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_FILE);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -ENOENT; goto fault;
    }
    fi->fh = (uint64_t) (uintptr_t) file_open(&at);
    if ((void *)(uintptr_t)fi->fh == NULL) ret = -ENOMEM;
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int create_empty(part_t *partition, offset_t *start) {
    buffer_t *sbuff;
    int ret = find_free_sector(partition, start, 0);
    if (ret) return ret;
    sbuff = buffer_new(*start);
    ret = sbuff->fault;
    buffer_done(sbuff);
    return ret;
}

static int extend_dir(resolv_t *at) {
    offset_t empty;
    buffer_t *sbuff;
    int ret;

    if (at->sectors >= 64) return -ENOSPC;

    empty = at->end_of_this_dir;
    ret = create_empty(at->partition, &empty);
    if (ret) return ret;
    at->empty_sector = empty;
    at->empty_offset = 0;

    sbuff = buffer_read(at->end_of_this_dir);
    ret = sbuff->fault;
    if (!ret) {
        set_dir_next(sbuff->buffer, empty);
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    return ret;
}

static offset_t get_treelink(const unsigned char *buffer) {
    unsigned char next_sector[4], c;
    const unsigned char *entry = buffer + 64;
    int i;

    for (i = 0; i < 4; i++) {
        entry -= 16;
        c = (entry[0] << 2) & 0xc0;
        c |= entry[4] & 0x30;
        c |= (entry[8] >> 2) & 0x0c;
        next_sector[i] = c | ((entry[12] >> 4) & 0x03);
    }
    return sector2offset(next_sector);
}

static void set_treelink(unsigned char *buffer, offset_t ezt) {
    unsigned char next_sector[4];
    unsigned char *entry = buffer + 64;
    int i, j;

    offset2sector(next_sector, ezt);

    for (j = 0; j < 4; j++) {
        unsigned int c = next_sector[j] << 4;
        for (i = 0; i < 4; i++) {
            entry -= 4;
            entry[0] = (entry[0] & ~0x30) | (c & 0x30);
            c >>= 2;
        }
    }
}

static void recurs_truncate(part_t *partition, offset_t itt, int level, unsigned int pos, unsigned int size, int modify) {
    buffer_t *sbuff;
    int i, j;
    int freep = 0;
    offset_t to_free[128 + 8], clean;

    if (level > 8) return;

    clean = null_sector;
    sbuff = buffer_read(itt);
    for (i = 0; i < 128; i++, pos += 512) {
        unsigned char *p = sbuff->buffer + i * 4;
        offset_t itt2;
        if (pos >= size) {
            if (modify) {
                if ((p[0] & 0x40) | p[1] | p[2] | p[3]) {
                    itt2 = sector2offset(p);
                    p[0] &= 0xb0;
                    p[1] = p[2] = p[3] = 0;
                    sbuff->dirty = 1;
                } else continue;
            } else itt2 = sector2offset(p);
            if (itt2 != null_sector) to_free[freep++] = itt2;
            continue;
        }
        if (pos == (size & ~511)) clean = sector2offset(p);
    }

    pos = (pos - 0xe000) << 3;
    for (i = 0; i < 512; i += 64, pos += 0x10000) {
        unsigned char *p = sbuff->buffer + (((i & 0x100) >> 2) + (i & 0x80) + ((i & 0x40) << 2));
        offset_t itt2;
        if (pos >= size) {
            for (j = 0; j < 16; j++) if (p[j * 4] & 0x30) break;
            if (j < 16) {
                itt2 = get_treelink(p);
                for (; j < 16; j++) p[j * 4] &= ~0x30;
                sbuff->dirty = 1;
            } else continue;
            if (itt2 == null_sector) continue;
            buffer_done(sbuff);
            to_free[freep++] = itt2;
            recurs_truncate(partition, itt2, level + 1, pos, size, 0);
            sbuff = buffer_read(itt);
            continue;
        }
        if (pos == (size & ~0xffff)) {
            itt2 = get_treelink(p);
            if (itt2 == null_sector) continue;
            buffer_done(sbuff);
            recurs_truncate(partition, itt2, level + 1, pos, size, 1);
            sbuff = buffer_read(itt);
        }
    }
    buffer_done(sbuff);

    for (i = 0; i < freep; i++) free_this_sector(partition, to_free[i]);

    if (clean != null_sector) {
        sbuff = buffer_read(clean);
        memset(sbuff->buffer + (size & 511), 0, 512 - (size & 511));
        sbuff->dirty = 1;
        buffer_done(sbuff);
    }
}

static int is_valid_name_type(const c64_t *name, c64_t *type) {
    int i;
    const char *ftype;

    for (i = 0; i < 16; i++) {
        c64_t c = name[i];
        if (c == '*' || c == '?' || c == ':' || c == '=' || c == ',') return -EINVAL;
        if (c == 0) break;
    }
    if (i == 0) return -EINVAL;

    for (i = 0; i < 3; i++) {
        c64_t c = type[i];
        if (c == '*' || c == '?' || c == ':' || c == '=' || c == ',' || c == '<' || c == '>' || c == ' ') return -EINVAL;
        if (c == 0) break;
    }
    if (i == 1) {
        switch (type[0]) {
        case 'P': ftype = "PRG"; break;
        case 'S': ftype = "SEQ"; break;
        case 'U': ftype = "USR"; break;
        default: return -EINVAL;
        }
        memcpy(type, ftype, 3);
    }
    return 0;
}

static void set_name_type(unsigned char *entry, const c64_t *name, const c64_t *type) {
    int i, j;
    for (j = i = 0; i < 16; i++) {
        entry[i] = name[j];
        if (name[j]) j++;
    }
    entry += 0x19;
    for (j = i = 0; i < 3; i++) {
        entry[i] = type[j];
        if (type[j]) j++;
    }
}

static int cfs011_create(const char *path, mode_t mode, struct fuse_file_info *fi) {
    offset_t empty;
    buffer_t *sbuff;
    resolv_t at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_FILE);
    if (ret < 0) goto fault;
    if (!ret) {
        if (mode & O_EXCL) {
            ret = -EEXIST; goto fault;
        }
        if (mode & O_TRUNC) {
            ret = -EEXIST; goto fault;
        }

        if (fi != NULL) {
            fi->fh = (uint64_t) (uintptr_t) file_open(&at);
            if ((void *)(uintptr_t)fi->fh == NULL) ret = -ENOMEM;
        }
        goto fault;
    }

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (!memcmp(at.type, "DEL", 3) || !memcmp(at.type, "REL", 3) ||
            !memcmp(at.type, "DIR", 3) || !memcmp(at.type, "LNK", 3)) {
        ret = -EINVAL; goto fault;
    }

    if (at.empty_sector == null_sector) {
        ret = extend_dir(&at);
        if (ret) goto fault;
    }

    empty = at.empty_sector;
    ret = create_empty(at.partition, &empty);
    if (ret) goto fault;

    sbuff = buffer_read(at.empty_sector);
    ret = sbuff->fault;
    if (!ret) {
        unsigned char *entry = sbuff->buffer + at.empty_offset;
        set_name_type(entry, at.name, at.type);
        memset(entry + 0x10, 0, 4);
        entry[0x14] &= ~0x80;
        offset2sector_insert(entry + 0x14, empty);
        entry[0x18] = 0xc0 | FTYPE_PRG | ((mode & 0700) >> 3);
        fill_in_time(entry + 0x1c, time(NULL));
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    if (ret) goto fault;
    if (fi != NULL) {
        at.root = empty;
        at.dir_sector = at.empty_sector;
        at.file_offset = at.empty_offset;
        fi->fh = (uint64_t) (uintptr_t) file_open(&at);
        if ((void *)(uintptr_t)fi->fh == NULL) ret = -ENOMEM;
    }
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_mknod(const char *path, mode_t mode, dev_t rdev) {
    (void) rdev;
    int ret;

    if (!(S_IFREG & mode)) {
        return -ENOSYS;
    }

    ret = cfs011_create(path, mode | O_EXCL, NULL);
    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static offset_t read_tree_search(offset_t itt, off_t offset, int *ret) {
    off_t m = offset / 512 / 128;
    off_t m2 = 0;
    unsigned int level = 0;

    while (m) {
        m--;
        m2 <<= 1; m2 |= m & 1; m >>= 1;
        m2 <<= 1; m2 |= m & 1; m >>= 1;
        m2 <<= 1; m2 |= m & 1; m >>= 1;
        level++;
    }

    while (level && itt != null_sector) {
        buffer_t *tbuff = buffer_read(itt);
        if (tbuff->fault) {
            *ret = tbuff->fault;
            itt = null_sector;
        } else {
            itt = get_treelink(tbuff->buffer + (m2 & 7) * 64);
            m2 >>= 3;
            level--;
        }
        buffer_done(tbuff);
    }
    return itt;
}

static int cfs011_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
    off_t last_offset = 1;
    off_t fsize;
    offset_t root, itt, itt2;
    buffer_t *sbuff, *tbuff;
    unsigned char *entry;
    int read, ret;
    file_t *file;
    (void) path;

    pthread_mutex_lock(&big_lock);

    read = 0;
    file = (file_t *) (uintptr_t) fi->fh;

    sbuff = buffer_read(file->dir_sector);
    ret = sbuff->fault;
    entry = sbuff->buffer + file->dir_offset;
    fsize = get_entry_size(entry);

    fsize = (fsize > offset) ? (fsize - offset) : 0;
    if (size > fsize) size = fsize;

    root = sector2offset(entry + 0x14);
    buffer_done(sbuff);

    goto start;
    while (size) {
        size_t size2;

        if (last_offset != (offset & ~0xffff)) {
        start:
            last_offset = offset & ~0xffff;
            itt = read_tree_search(root, offset, &ret);
        }
   
        size2 = size;
        if (size2 > 512 - (offset & 511)) {
            size2 = 512 - (offset & 511);
        }

        itt2 = itt;
        if (itt2 != null_sector) {
            tbuff = buffer_read(itt);
            if (tbuff->fault) {
                ret = tbuff->fault;
                itt2 = null_sector;
            } else {
                itt2 = sector2offset(tbuff->buffer + ((offset >> 7) & 0x1fc));
            }
            buffer_done(tbuff);
        }

        if (itt2 == null_sector) {
            memset(buf, 0, size2);
        } else {
            int i;
            unsigned char *buf2;
            buffer_t *b = buffer_find(itt2);
            if (size2 == 512) {
                unsigned char buf3[512];
                if (b->offset != itt2) {
                    int ret2 = read_sector(buf3, itt2);
                    if (ret2) ret = ret2;
                    buf2 = buf3;
                } else {
                    if (b->fault) ret = b->fault;
                    buf2 = b->buffer;
                }
                for (i = 0; i < 128; i++) {
                    buf[i] = buf2[i * 4];
                    buf[i + 128] = buf2[i * 4 + 1];
                    buf[i + 256] = buf2[i * 4 + 2];
                    buf[i + 384] = buf2[i * 4 + 3];
                }
                if (b->offset == itt2) buffer_done(b);
            } else {
                unsigned int j;
                if (b->offset != itt2) {
                    buffer_flush(b);
                    b->fault = read_sector(b->buffer, itt2);
                    b->offset = itt2;
                }
                if (b->fault) ret = b->fault;
                buf2 = b->buffer;
                i = (offset & 511);
                for (j = 0; j < size2; j++, i++) buf[j] = buf2[(i >> 7) + ((i << 2) & 0x1fc)];
                buffer_done(b);
            }
        }

        buf += size2;
        read += size2;
        offset += size2;
        size -= size2;
    }
    ret = read;
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
    off_t last_offset = 1;
    offset_t itt = null_sector, empty;
    buffer_t *sbuff, *tbuff;
    unsigned char *entry;
    int write, ret, ret2;
    file_t *file;
    (void) path;

    pthread_mutex_lock(&big_lock);

    write = 0;
    file = (file_t *) (uintptr_t) fi->fh;

    while (size) {
        unsigned int i, j;
        size_t size2 = size;
        int holecheck;

        if (size2 > 512 - (offset & 511)) {
            size2 = 512 - (offset & 511);
        }
        holecheck = (size2 == 512);

        if (last_offset != (offset & ~0xffff)) {
            off_t m = offset / 512 / 128;
            off_t m2 = 0;
            unsigned int level = 0;

            while (m) {
                m--;
                m2 <<= 1; m2 |= m & 1; m >>= 1;
                m2 <<= 1; m2 |= m & 1; m >>= 1;
                m2 <<= 1; m2 |= m & 1; m >>= 1;
                level++;
            }

            itt = file->treeroot;

            while (level) {
                offset_t itt2;

                tbuff = buffer_read(itt);
                if (tbuff->fault) {
                    ret = tbuff->fault;
                    buffer_done(tbuff);
                    goto hole;
                }
                itt2 = get_treelink(tbuff->buffer + (m2 & 7) * 64);
                buffer_done(tbuff);
                if (itt2 == null_sector) {// it does not exists yet! (empty sector)
                    if (holecheck) {
                        for (i = 0; i < size2; i++) {
                            if (buf[i] != 0) break;
                        }
                        if (i == size2) goto hole;
                        holecheck = 0;
                    }
                    empty = file->lasttree;
                    ret2 = create_empty(file->partition, &empty);
                    if (ret2) {
                        ret = ret2;
                        goto hole;
                    }

                    tbuff = buffer_read(itt);
                    set_treelink(tbuff->buffer + (m2 & 7) * 64, empty);
                    tbuff->dirty = 1;
                    buffer_done(tbuff);
                    itt = empty;
                } else {
                    itt = itt2;
                }

                m2 >>= 3;
                level--;
            }
            last_offset = offset & ~0xffff;
        }

        tbuff = buffer_read(itt);
        empty = sector2offset(tbuff->buffer + ((offset >> 7) & 0x1fc));
        buffer_done(tbuff);

        if (empty == null_sector) {
            if (holecheck) {
                for (i = 0; i < size2; i++) {
                    if (buf[i] != 0) break;
                }
                if (i == size2) goto hole;
                holecheck = 0;
            }
            empty = file->lastblock;
            ret2 = find_free_sector(file->partition, &empty, 1);
            if (ret2) {
                ret = ret2;
                goto hole;
            }

            tbuff = buffer_read(itt);
            offset2sector_insert(tbuff->buffer + ((offset >> 7) & 0x1fc), empty);
            tbuff->dirty = 1;
            buffer_done(tbuff);

            if (size2 != 512) sbuff = buffer_new(empty);
        } else {
            if (size2 != 512) sbuff = buffer_read(empty);
        }
        if (size2 == 512) {
            unsigned char buf3[512], *buf2;
            sbuff = buffer_find(empty);
            buf2 = (sbuff->offset != empty) ? buf3 : sbuff->buffer;
            for (i = 0; i < 128; i++) {
                 buf2[i * 4] = buf[i];
                 buf2[i * 4 + 1] = buf[i + 128];
                 buf2[i * 4 + 2] = buf[i + 256];
                 buf2[i * 4 + 3] = buf[i + 384];
            }
            if (sbuff->offset != empty) {
                ret2 = write_sector(buf3, empty);
                if (ret2) ret = ret2;
            } else {
                sbuff->dirty = 1;
                buffer_done(sbuff);
            }
        } else {
            i = (offset & 511);
            for (j = 0; j < size2; j++, i++) sbuff->buffer[(i >> 7) + ((i << 2) & 0x1fc)] = buf[j];
            sbuff->dirty = 1; 
            buffer_done(sbuff);
        }
        file->lastblock = empty;
    hole:
        write += size2;
        buf += size2;
        offset += size2;
        size -= size2;
    }

    sbuff = buffer_read(file->dir_sector);
    entry = sbuff->buffer + file->dir_offset;

    if (get_entry_size(entry) < offset) {
        set_entry_size(entry, offset);
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    ret = write;

    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int change_attr(mode_t mode, offset_t itt, unsigned int offset) {
    buffer_t *sbuff = buffer_read(itt);
    int ret = sbuff->fault;

    if (!ret) {
        unsigned char *entry = sbuff->buffer + offset;
        unsigned char attr = (entry[0x18] & 0xc7) | ((mode & 0700) >> 3);
        if (entry[0x18] != attr) {
            entry[0x18] = attr;
            sbuff->dirty = 1;
        }
        ret = (attr & 0x87) == 0x83;
    }
    buffer_done(sbuff);
    return ret;
}

static int cfs011_chmod(const char *path, mode_t mode) {
    resolv_t at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.root == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    if (at.root != at.partition->rootdir) {
        ret = change_attr(mode, at.dir_sector, at.file_offset);
        if (ret > 0) ret = change_attr(mode, at.root, 0);
    }
fault:
    //ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_chown(const char *path, uid_t uid, gid_t gid) {
    resolv_t at;
    (void) uid;
    (void) gid;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int change_time(const unsigned char *newtime, offset_t itt, unsigned int offset) {
    buffer_t *sbuff = buffer_read(itt);
    int ret = sbuff->fault;

    if (!ret) {
        unsigned char *entry = sbuff->buffer + offset;
        ret = (entry[0x18] & 0x87) == 0x83;
        entry += 0x1c;
        if (memcmp(entry, newtime, 4) != 0) {
            memcpy(entry, newtime, 4);
            sbuff->dirty = 1;
        }
    }
    buffer_done(sbuff);
    return ret;
}

static int cfs011_utimens(const char *path, const struct timespec tv[2]) {
    resolv_t at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.root == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    if (at.root != at.partition->rootdir) {
        unsigned char newtime[4];
        fill_in_time(newtime, tv[1].tv_sec);
        ret = change_time(newtime, at.dir_sector, at.file_offset);
        if (ret > 0) ret = change_time(newtime, at.root, 0);
    }
fault:
    //ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_symlink(const char *from, const char *to) {
    offset_t empty;
    resolv_t at;
    buffer_t *sbuff;
    unsigned char *entry;
    int i;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, to, RESOLV_LINK);
    if (!ret) ret = -EEXIST;
    if (ret < 0) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (memcmp(at.type, "LNK", 3)) {
        ret = -EINVAL; goto fault;
    }

    if (at.empty_sector == null_sector) {
        ret = extend_dir(&at);
        if (ret) goto fault;
    }

    empty = at.empty_sector;
    ret = create_empty(at.partition, &empty);
    if (ret) goto fault;

    sbuff = buffer_read(empty);
    ret = sbuff->fault;
    if (ret) goto fault;
    to_petscii(sbuff->buffer, from, strlen(from), 511);
    i = strlen((const char *)sbuff->buffer);
    sbuff->dirty = 1;
    buffer_done(sbuff);

    sbuff = buffer_read(at.empty_sector);
    entry = sbuff->buffer + at.empty_offset;
    set_name_type(entry, at.name, (const c64_t *)"LNK");
    entry[0x10] = i;
    entry[0x11] = i >> 8;
    entry[0x12] = 0;
    entry[0x13] = 0;
    entry[0x14] &= ~0x80;
    offset2sector_insert(entry + 0x14, empty);
    entry[0x18] = 0xf8 | FTYPE_LNK;
    fill_in_time(entry + 0x1c, time(NULL));
    sbuff->dirty = 1;
    buffer_done(sbuff);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_rename(const char *from, const char *to) {
    resolv_t at, at2;
    buffer_t *sbuff;
    unsigned char *entry;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at2, from, RESOLV_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at2.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at2.root == at2.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    ret = resolv_path(&at, to, (at2.attr & S_IFDIR) ? RESOLV_DIR : (at2.attr & S_IFLNK) ? RESOLV_LINK : RESOLV_FILE);
    if (ret < 0) goto fault;
    if (!ret) {
        if (at.partition != at2.partition || at.dir_sector != at2.dir_sector || at.file_offset != at2.file_offset) ret = -EEXIST; 
        goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    if (at.partition != at2.partition) {
        ret = -EXDEV; goto fault;
    }
    if (at.this_dir != at2.this_dir) {
        ret = -EXDEV; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (memcmp(at.type, at2.type, 3) && (!memcmp(at.type, "DEL", 3) || !memcmp(at.type, "REL", 3) ||
        !memcmp(at.type, "DIR", 3) || !memcmp(at.type, "LNK", 3))) {
        ret = -EINVAL; goto fault;
    }

    sbuff = buffer_read(at2.dir_sector);
    at.empty_sector = at2.dir_sector;
    at.empty_offset = at2.file_offset;

    entry = sbuff->buffer + at.empty_offset;
    set_name_type(entry, at.name, (entry[0x18] & FTYPE_MASK) == FTYPE_PRG ? at.type : at2.type);
    sbuff->dirty = 1;
    buffer_done(sbuff);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_mkdir(const char *path, mode_t mode) {
    offset_t empty;
    resolv_t at;
    buffer_t *sbuff;
    unsigned char *entry;
    int i, j;
    time_t ee;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_DIR);
    if (!ret) ret = -EEXIST;
    if (ret < 0) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (memcmp(at.type, "DIR", 3)) {
        ret = -EINVAL; goto fault;
    }

    if (at.empty_sector == null_sector) {
        ret = extend_dir(&at);
        if (ret) goto fault;
    }

    empty = at.empty_sector;
    ret = find_free_sector(at.partition, &empty, 0);
    if (ret) goto fault;

    sbuff = buffer_new(empty);
    entry = sbuff->buffer;
    for (j = i = 0; i < 16; i++) {
        if (at.name[j]) entry[i] = at.name[j++];
        else entry[i] = 0x20;
    }
    offset2sector(entry + 0x10, empty);
    offset2sector(entry + 0x14, at.this_dir);
    entry[0x18] = 0x40 | FTYPE_DIR | ((mode & 0700) >> 3);
    memcpy(entry + 0x19, "DIR", 3);
    ee = time(NULL);
    fill_in_time(entry + 0x1c, ee);
    sbuff->dirty = 1;
    buffer_done(sbuff);

    sbuff = buffer_read(at.empty_sector);
    entry = sbuff->buffer + at.empty_offset;
    set_name_type(entry, at.name, (const c64_t *)"DIR");
    memset(entry + 0x10, 0, 4);
    entry[0x14] &= ~0x80;
    offset2sector_insert(entry + 0x14, empty);
    entry[0x18] = 0xc0 | FTYPE_DIR | ((mode & 0700) >> 3);
    fill_in_time(entry + 0x1c, ee);
    sbuff->dirty = 1;
    buffer_done(sbuff);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_ftruncate(const char *path, off_t size, struct fuse_file_info *fi) {
    resolv_t at;
    buffer_t *sbuff;
    unsigned char *entry;
    unsigned int size2;
    int ret = 0;

    pthread_mutex_lock(&big_lock);

    if (fi == NULL) {
        ret = resolv_path(&at, path, RESOLV_FILE);
        if (ret > 0) ret = -ENOENT;
        if (ret) goto fault;

        if (at.partition == NULL) {
            ret = -ENOENT; goto fault;
        }
    } else {
        file_t *file = (file_t *) (uintptr_t) fi->fh;
        at.dir_sector = file->dir_sector;
        at.file_offset = file->dir_offset;
        at.partition = file->partition;
        at.root = file->treeroot;
    }

    sbuff = buffer_read(at.dir_sector);
    entry = sbuff->buffer + at.file_offset;
    size2 = get_entry_size(entry);

    switch (entry[0x18] & FTYPE_MASK) {
    case FTYPE_REL:
        if (size > 0xffffff) {
            ret = -EFBIG; goto fault;
        }
        /* fall through */
    case FTYPE_PRG:
        if (size <= size2) {
            recurs_truncate(at.partition, at.root, 0, 0, size, size < 65536);
        }
        break;
    case FTYPE_LNK:
        if (size > 0x200) {
            ret = -EFBIG; goto fault;
        }
        break;
    }

    if (size != size2) {
        set_entry_size(entry, size);
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_truncate(const char *path, off_t size) {
    int ret = cfs011_ftruncate(path, size, NULL);
    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int is_dir_empty(const unsigned char *buffer, int except) {
    int i;
    for (i = 0; i < 512; i += 32) {
        const unsigned char *entry = buffer + i;
        unsigned char attr = entry[0x18];
        if ((attr & 0x80) != 0x00 && (attr & 0x87) != 0x03) {
            if (i != except) return 0;
        }
    }
    return 1;
}

static int delete_entry(const resolv_t *at) {
    int ret;
    buffer_t *sbuff = buffer_read(at->dir_sector);
    ret = sbuff->fault;
    if (!ret) ret = sbuff->buffer[at->file_offset + 0x18] & 0x87;

    if (ret >= 0 && at->end_of_this_dir != null_sector) {
        if (is_dir_empty(sbuff->buffer, at->file_offset)) {
            offset_t itt = get_dir_next(sbuff->buffer);
            buffer_done(sbuff);
            sbuff = buffer_read(at->end_of_this_dir);
            if (sbuff->fault) {
                ret = sbuff->fault;
                buffer_done(sbuff);
                return ret;
            } 
            set_dir_next(sbuff->buffer, itt);
            sbuff->dirty = 1;
            buffer_done(sbuff);
            free_this_sector(at->partition, at->dir_sector);
            return ret;
        }
    }
    if (ret >= 0) {
        sbuff->buffer[at->file_offset + 0x18] &= 0x78;
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    return ret;
}

static int cfs011_unlink(const char *path) {
    resolv_t at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_FILE);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -ENOENT; goto fault;
    }
    if (file_lookup(&at) != NULL) {
        ret = -EBUSY; goto fault;
    }

    ret = delete_entry(&at);
    if (ret < 0) goto fault;

    switch (ret) {
    case 0x80: break;
    case 0x81:
    case 0x82:
        recurs_truncate(at.partition, at.root, 0, 0, 0, 0);
        free_this_sector(at.partition, at.root);
        break;
    case 0x84: free_this_sector(at.partition, at.root); break;
    }
    ret = 0;
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int rmdir_walk(const resolv_t *at, int delete) {
    int ret = 0;
    offset_t itt, next;

    for (itt = at->root; itt != null_sector; itt = next) {
        buffer_t *sbuff = buffer_read(itt);
        ret = sbuff->fault;
        if (ret) {
            buffer_done(sbuff);
            break;
        }
        if (!delete && !is_dir_empty(sbuff->buffer, -1)) {
            buffer_done(sbuff);
            return -ENOTEMPTY;
        }
        next = get_dir_next(sbuff->buffer);
        buffer_done(sbuff);
        if (delete) free_this_sector(at->partition, itt);
    }
    return ret;
}

static int cfs011_rmdir(const char *path) {
    resolv_t at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_DIR);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.root == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = rmdir_walk(&at, 0);
    if (ret) goto fault;

    ret = delete_entry(&at);
    if (ret < 0) goto fault;

    ret = rmdir_walk(&at, 1);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_statfs(const char *path, struct statvfs *stbuf) {
    resolv_t at;
    static int fs_free_updated = 0;
    int i, j;
    offset_t itt;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (!fs_free_updated) {
        for (i = 0; i < 16; i++) {
            off_t freeblk;
            part_t *partition = &partitions[i];
            if (!partition->valid) continue;
            partition->size = partition->end - partition->start;
            freeblk = 0;

            itt = partition->bam;
            do {
                buffer_t *sbuff = buffer_read(itt);
                if (sbuff->fault) {
                    ret = sbuff->fault;
                } else {
                    unsigned char f;
                    for (j = 0; j < 512; j++) {
                        f = sbuff->buffer[j];
                        if (f == 0xff) {
                            freeblk += 8;
                            continue;
                        }
                        while (f) {
                            freeblk += f & 1;
                            f >>= 1;
                        }
                    }
                    f = sbuff->buffer[0];
                    if (f & 0x80) freeblk--;
                    if (f & 0x40) freeblk--;
                }
                buffer_done(sbuff);
                itt += 4096;
                partition->size -= 2;
            } while (itt < partition->end);
            partition->free = freeblk;
        }

        fs_free_updated = 1;
    }
    if (at.partition != NULL) {
        stbuf->f_bfree = at.partition->free;
        stbuf->f_blocks = at.partition->size;
    } else {
        stbuf->f_bfree = 0;
        stbuf->f_blocks = 0;
        for (i = 0; i < 16; i++) {
            part_t *partition = &partitions[i];
            if (!partition->valid) continue;
            if (partition->rootdir == at.root) {
                stbuf->f_bfree = partition->free;
                stbuf->f_blocks = partition->size;
                break;
            }
            stbuf->f_bfree += partition->free;
            stbuf->f_blocks += partition->size;
        }
    }
    stbuf->f_bavail = stbuf->f_bfree;
    //stbuf->f_type = 0x64;
    stbuf->f_bsize = 512;
    stbuf->f_frsize = 512;
    stbuf->f_files = 0;
    stbuf->f_ffree = 0;
#ifdef __CYGWIN__
    stbuf->f_namemax = 40;
#else
    stbuf->f_namemax = 20;
#endif
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
    (void) path;
    (void) isdatasync;
    (void) fi;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(0);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_readlink(const char *path, char *out, size_t size) {
    resolv_t at;
    buffer_t *sbuff;
    char buf[512*MB_CUR_MAX + 1];
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolv_path(&at, path, RESOLV_LINK);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;
    
    if (at.partition == NULL) {
        ret = -ENOENT; goto fault;
    }

    sbuff = buffer_read(at.root);
    ret = sbuff->fault;
    to_ascii(buf, sbuff->buffer, 1);
    buffer_done(sbuff);
    memcpy(out, buf, size);
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_flush(const char *path, struct fuse_file_info *fi) {
    (void) path;
    (void) fi;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(0);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_release(const char *path, struct fuse_file_info *fi) {
    (void) path;
    file_t *file = (file_t *) (uintptr_t) fi->fh;

    file->refcount--;
    if (file->refcount == 0) free((file_t *) (uintptr_t) fi->fh);
    return 0;
}

static void *cfs011_init(struct fuse_conn_info *conn) {
    int i, fault;
    buffer_t *sbuff;
    offset_t backup_table;
    struct tm epoch;

    stat_default.st_blksize = 512;
    stat_default.st_mtime = time(NULL);
    stat_default.st_atime = stat_default.st_mtime;
    stat_default.st_ctime = stat_default.st_mtime;

    localtime_r(&stat_default.st_mtime, &epoch);
    epoch.tm_sec = 0;
    epoch.tm_min = 0;
    epoch.tm_hour = 23;
    epoch.tm_mday = 29;
    epoch.tm_mon = 10;
    epoch.tm_year = 79;
    epoch.tm_wday = 0;
    epoch.tm_yday = 0;
    epoch.tm_isdst = 0;
    epoch_time = mktime(&epoch);

#ifdef FUSE_CAP_BIG_WRITES
    conn->want |= FUSE_CAP_BIG_WRITES;
#endif

    for (i = 0; i < CACHE_SIZE; i++) {
        buffers[i].offset = -1;
        buffers[i].next = i + 1;
    }
    buffer_head = 0;

    sbuff = buffer_read(0);
    geometry_head = (sbuff->buffer[0x04] & 0x4f) + 1;
    geometry_cylinder = (sbuff->buffer[0x05] << 8) + sbuff->buffer[0x06];
    geometry_sector = sbuff->buffer[0x07];
    fault = (memcmp(sbuff->buffer + 0x08, "C64 CFS V 0.11B ", 16) != 0) || sbuff->fault;
    fault |= !(geometry_head & 0x40) && (geometry_head == 0 || geometry_sector == 0);
    buffer_done(sbuff);
    if (fault) return NULL;
   
    partition_table = sector2offset(sbuff->buffer + 0x18);
    backup_table = sector2offset(sbuff->buffer + 0x1c);
    sbuff = buffer_read(partition_table);
    if (sbuff->fault) {
        buffer_done(sbuff);
        partition_table = backup_table;
        sbuff = buffer_read(partition_table);
    }
    if (!sbuff->fault) {
        for (i = 0; i < 16; i++) {
            unsigned char *entry = sbuff->buffer + i * 32;
            part_t *partition = &partitions[i];
            partition->valid = (entry[0x10] & 0x80) && (entry[0x14] & 0xB0) == 0x10;
            partition->rootdir = sector2offset(entry + 0x1c);
            partition->start = sector2offset(entry + 0x10);
            partition->end = sector2offset(entry + 0x14);
            partition->deldir = sector2offset(entry + 0x18);
            partition->bam = (entry[0x1c] >> 7) + partition->start;
            memcpy(partition->name, entry, 16);
            partition->name[16] = 0;
            if (partition->rootdir < partition->start || partition->rootdir >= partition->end) partition->valid = 0;
        }
    }
    buffer_done(sbuff);
    return NULL;
}

static void cfs011_destroy(void *userdata) {
    (void) userdata;
    buffers_sync(0);
    close(f1);
}

static struct fuse_operations cfs011_oper = {
    .getattr	= cfs011_getattr,
    .readlink	= cfs011_readlink,
    .mknod	= cfs011_mknod,
    .mkdir	= cfs011_mkdir,
    .unlink	= cfs011_unlink,
    .rmdir	= cfs011_rmdir,
    .symlink	= cfs011_symlink,
    .rename	= cfs011_rename,
    .link       = NULL,
    .chmod	= cfs011_chmod,
    .chown	= cfs011_chown,
    .truncate	= cfs011_truncate,
    .utimens	= cfs011_utimens,
    .open	= cfs011_open,
    .read	= cfs011_read,
    .write	= cfs011_write,
    .statfs	= cfs011_statfs,
    .flush      = cfs011_flush,
    .release	= cfs011_release,
    .fsync	= cfs011_fsync,
    .setxattr   = NULL,
    .getxattr   = NULL,
    .listxattr  = NULL,
    .removexattr= NULL,
    .opendir    = NULL,
    .readdir	= cfs011_readdir,
    .releasedir = NULL,
    .fsyncdir   = NULL,
    .init       = cfs011_init,
    .destroy    = cfs011_destroy,
    .create	= cfs011_create,
    .ftruncate	= cfs011_ftruncate,
    .fgetattr	= cfs011_fgetattr,
};

int main(int argc, char *argv[]) {
    setlocale(LC_ALL, "");

    puts("FUSECFS011 v1.9 (C) 2005-2017 Kajtar Zsolt (Soci/Singular)");
    if (argc < 2) {
        printf("Usage: %s <device> <mount_point> [<FUSE library options>]\nUnmount: fusermount -u <mount_point>\n", argv[0]);
	return 1;
    }

    f1 = open(argv[1], O_RDWR | O_LARGEFILE);
    if (f1 < 0) {
        printf("Cannot open image file \"%s\"\n", argv[1]);
	return 1;
    }
    argv[1] = argv[0];

    return fuse_main(argc - 1, argv + 1, &cfs011_oper, NULL);
}
